/***
  This file is part of PulseAudio.

  Copyright 2018 Pali Rohár <pali.rohar@gmail.com>

  PulseAudio is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2.1 of the
  License, or (at your option) any later version.

  PulseAudio is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with PulseAudio; if not, see <http://www.gnu.org/licenses/>.
***/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <pulsecore/log.h>
#include <pulsecore/macro.h>
#include <pulsecore/once.h>
#include <pulse/sample.h>

#include <openaptx.h>

#include "a2dp-codecs.h"
#include "pa-a2dp-codec.h"

static size_t pa_aptx_fill_capabilities(uint8_t *capabilities_buffer, size_t capabilities_size) {
    a2dp_aptx_t *capabilities = (a2dp_aptx_t *) capabilities_buffer;

    if (capabilities_size < sizeof(*capabilities)) {
        pa_log_error("Invalid size of capabilities buffer");
        return 0;
    }

    pa_zero(*capabilities);

    capabilities->info.vendor_id = A2DP_VENDOR_APT_LIC_LTD;
    capabilities->info.codec_id = APT_LIC_LTD_CODEC_APTX;
    capabilities->channel_mode = APTX_CHANNEL_MODE_STEREO;
    capabilities->frequency = APTX_SAMPLING_FREQ_16000 | APTX_SAMPLING_FREQ_32000 | APTX_SAMPLING_FREQ_44100 |
                             APTX_SAMPLING_FREQ_48000;

    return sizeof(*capabilities);
}

static bool pa_aptx_validate_configuration(const uint8_t *config_buffer, size_t config_size) {
    a2dp_aptx_t *config = (a2dp_aptx_t *) config_buffer;

    if (config_size != sizeof(*config)) {
        pa_log_error("Invalid size of config buffer");
        return false;
    }

    if (config->info.vendor_id != A2DP_VENDOR_APT_LIC_LTD || config->info.codec_id != APT_LIC_LTD_CODEC_APTX) {
        pa_log_error("Invalid vendor codec information in configuration");
        return false;
    }

    if (config->frequency != APTX_SAMPLING_FREQ_16000 && config->frequency != APTX_SAMPLING_FREQ_32000 &&
        config->frequency != APTX_SAMPLING_FREQ_44100 && config->frequency != APTX_SAMPLING_FREQ_48000) {
        pa_log_error("Invalid sampling frequency in configuration");
        return false;
    }

    if (config->channel_mode != APTX_CHANNEL_MODE_STEREO) {
        pa_log_error("Invalid channel mode in configuration");
        return false;
    }

    return true;
}

static size_t pa_aptx_select_configuration(const pa_sample_spec *sample_spec, const uint8_t *capabilities_buffer, size_t capabilities_size, uint8_t *config_buffer, size_t config_size) {
    a2dp_aptx_t *config = (a2dp_aptx_t *) config_buffer;
    a2dp_aptx_t *capabilities = (a2dp_aptx_t *) capabilities_buffer;
    int i;

    static const struct {
        uint32_t rate;
        uint8_t cap;
    } freq_table[] = {
        { 16000U, APTX_SAMPLING_FREQ_16000 },
        { 32000U, APTX_SAMPLING_FREQ_32000 },
        { 44100U, APTX_SAMPLING_FREQ_44100 },
        { 48000U, APTX_SAMPLING_FREQ_48000 }
    };

    if (capabilities_size != sizeof(*capabilities)) {
        pa_log_error("Invalid size of capabilities buffer");
        return 0;
    }

    if (config_size < sizeof(*config)) {
        pa_log_error("Invalid size of config buffer");
        return 0;
    }

    pa_zero(*config);

    if (capabilities->info.vendor_id != A2DP_VENDOR_APT_LIC_LTD || capabilities->info.codec_id != APT_LIC_LTD_CODEC_APTX) {
        pa_log_error("No supported vendor codec information");
        return 0;
    }

    config->info.vendor_id = A2DP_VENDOR_APT_LIC_LTD;
    config->info.codec_id = APT_LIC_LTD_CODEC_APTX;

    if (sample_spec->channels != 2 || !(capabilities->channel_mode & APTX_CHANNEL_MODE_STEREO)) {
        pa_log_error("No supported channel modes");
        return 0;
    }

    config->channel_mode = APTX_CHANNEL_MODE_STEREO;

    /* Find the lowest freq that is at least as high as the requested sampling rate */
    for (i = 0; (unsigned) i < PA_ELEMENTSOF(freq_table); i++) {
        if (freq_table[i].rate >= sample_spec->rate && (capabilities->frequency & freq_table[i].cap)) {
            config->frequency = freq_table[i].cap;
            break;
        }
    }

    if ((unsigned) i == PA_ELEMENTSOF(freq_table)) {
        for (--i; i >= 0; i--) {
            if (capabilities->frequency & freq_table[i].cap) {
                config->frequency = freq_table[i].cap;
                break;
            }
        }

        if (i < 0) {
            pa_log_error("Not suitable sample rate");
            return 0;
        }
    }

    return sizeof(*config);
}

static void pa_aptx_init(void **info_ptr, pa_sample_spec *sample_spec, bool is_a2dp_sink, const uint8_t *config_buffer, size_t config_size) {
    a2dp_aptx_t *config = (a2dp_aptx_t *) config_buffer;

    pa_assert(config_size == sizeof(*config));

    sample_spec->format = PA_SAMPLE_S24LE;

    if (*info_ptr)
        aptx_reset((struct aptx_context *) *info_ptr);
    else
        *info_ptr = (void *)aptx_init(0);

    switch (config->frequency) {
        case APTX_SAMPLING_FREQ_16000:
            sample_spec->rate = 16000U;
            break;
        case APTX_SAMPLING_FREQ_32000:
            sample_spec->rate = 32000U;
            break;
        case APTX_SAMPLING_FREQ_44100:
            sample_spec->rate = 44100U;
            break;
        case APTX_SAMPLING_FREQ_48000:
            sample_spec->rate = 48000U;
            break;
        default:
            pa_assert_not_reached();
    }

    switch (config->channel_mode) {
        case APTX_CHANNEL_MODE_STEREO:
            sample_spec->channels = 2;
            break;
        default:
            pa_assert_not_reached();
    }
}

static void pa_aptx_finish(void **info_ptr) {
    struct aptx_context *aptx_c = (struct aptx_context *) *info_ptr;

    if (aptx_c) {
        aptx_finish(aptx_c);
        *info_ptr = NULL;
    }
}

static void pa_aptx_setup(void **info_ptr) {
}

static void pa_aptx_fill_blocksize(void **info_ptr, size_t read_link_mtu, size_t write_link_mtu, size_t *read_block_size, size_t *write_block_size) {
    *write_block_size = (write_link_mtu/(8*4)) * 8*4*6;
    *read_block_size = (read_link_mtu/(8*4)) * 8*4*6;
}

static bool pa_aptx_fix_latency(void **info_ptr) {
    return false;
}

static size_t pa_aptx_encode(void **info_ptr, uint32_t timestamp, const uint8_t *input_buffer, size_t input_size, uint8_t *output_buffer, size_t output_size) {
    struct aptx_context *aptx_c = (struct aptx_context *) *info_ptr;
    uint8_t *d;
    const uint8_t *p;
    size_t to_write, to_encode;

    p = input_buffer;
    to_encode = input_size;

    d = output_buffer;
    to_write = output_size;

    while (PA_LIKELY(to_encode > 0 && to_write > 0)) {
        size_t written;
        size_t encoded;
        encoded = aptx_encode(aptx_c, p, to_encode, d, to_write, &written);

        if (PA_UNLIKELY(encoded == 0)) {
            pa_log_error("aptX encoding error");
            return 0;
        }

        pa_assert_fp((size_t) encoded <= to_encode);
        pa_assert_fp((size_t) written <= to_write);

        p += encoded;
        to_encode -= encoded;

        d += written;
        to_write -= written;
    }

    pa_assert(to_encode == 0);

    PA_ONCE_BEGIN {
        pa_log_debug("Using aptX encoder implementation: libopenaptx from https://github.com/pali/libopenaptx");
    } PA_ONCE_END;

    return d - output_buffer;
}

static size_t pa_aptx_decode(void **info_ptr, const uint8_t *input_buffer, size_t input_size, uint8_t *output_buffer, size_t output_size, size_t *processed) {
    struct aptx_context *aptx_c = (struct aptx_context *) *info_ptr;

    const uint8_t *p;
    uint8_t *d;
    size_t to_write, to_decode;

    p = input_buffer;
    to_decode = input_size;

    d = output_buffer;
    to_write = output_size;

    while (PA_LIKELY(to_decode > 0)) {
        size_t written;
        size_t decoded;

        decoded = aptx_decode(aptx_c, p, to_decode, d, to_write, &written);

        if (PA_UNLIKELY(decoded == 0)) {
            pa_log_error("aptX decoding error");
            *processed = p - input_buffer;
            return 0;
        }

        pa_assert_fp((size_t) decoded <= to_decode);

        p += decoded;
        to_decode -= decoded;

        d += written;
        to_write -= written;
    }

    *processed = p - input_buffer;
    return d - output_buffer;
}

const pa_a2dp_codec_t pa_a2dp_codec_aptx = {
    .codec_id = A2DP_CODEC_VENDOR,
    .fill_capabilities = pa_aptx_fill_capabilities,
    .validate_configuration = pa_aptx_validate_configuration,
    .select_configuration = pa_aptx_select_configuration,
    .init = pa_aptx_init,
    .finish = pa_aptx_finish,
    .setup = pa_aptx_setup,
    .fill_blocksize = pa_aptx_fill_blocksize,
    .fix_latency = pa_aptx_fix_latency,
    .encode = pa_aptx_encode,
    .decode = pa_aptx_decode,
};
